////////////////////////////////////////
///@file rule_sample.c
///
///@brief pag接口五元组规则操作示例代码，主线程可以读入规则文件，收包线程可以加载或删除动态规则
///
/// 修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-10
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <netinet/ip.h> //struct ip

//本地路径下包含文件
#include "libpag.h"

///// 宏定义 ///////

#define PKT_HEADER_LEN 50 ///<报文附加包头的长度sizeof(struct pkt_header)
#define MAX_THREADS_NUM 4  ///<线程数 

////////// 数据 /////////

int glob_loop_flag_main;
int glob_loop_flag_thread;
int glob_print_pkt_flag;
int glob_rule_num;
int glob_rule_operate;

///计数器，每个线程的计数器内存空间分开，以避免cache抖动
ulong glob_pkts[MAX_THREADS_NUM][16];
ulong glob_bytes[MAX_THREADS_NUM][16];

ulong glob_hits[MAX_THREADS_NUM][16];

///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s [-f rule_file] [-r rule_num][-t operate_type] [-p]\n", cmd);
    printf("  -f rule config file to load, default = no static rules\n");
    printf("  -r dynamic rule num, default = 0, no dynamic rules\n");
    printf("  -t dynamic rule operate type, default = 0, add = 0, del = 1\n");
    printf("  -p print pkts and headers\n");
    printf("  -s rule ratio file to load\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_thread(void *ptr)
{
    int i;
    ulong old_pkts[MAX_THREADS_NUM];
    ulong old_bytes[MAX_THREADS_NUM];
    ulong old_hits[MAX_THREADS_NUM];

    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        old_pkts[i] = 0;
        old_bytes[i] = 0;
        old_hits[i] = 0;
    }
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        printf("\nCurrent speed----- ");
        printf("\n");        
        for (i = 0; i < MAX_THREADS_NUM; i++) 
        {
            printf("\t Thread[%d]:\t %8lu pps, %8lu M bps, %8lu hits\n", i, 
            glob_pkts[i][0] - old_pkts[i], 
            (glob_bytes[i][0] - old_bytes[i]) >> 17,
            glob_hits[i][0] - old_hits[i]);
            old_pkts[i] = glob_pkts[i][0];
            old_bytes[i] = glob_bytes[i][0];
            old_hits[i] = glob_hits[i][0];
        }
    }
}

///@brief 打印报文
///@param pkt 指向报文的指针
///@param len 报文长度
static inline void print_pkt(void *pkt, int len)
{
    int i;
    __u8 *p8;
    static int pkt_print;
    int total_length, frame_length;
    __u8 frame_type;
 
    pkt_print++;
    printf("\n====  pkt:%d  ==== len = %d\n", pkt_print , len);
    
    ///找到报文附加头, 打印长度包括附加头
    p8 = (__u8 *)pkt;
    frame_type = pag_get_frame_type(pkt);
    frame_length = pag_get_frame_length(pkt);
    
    if (frame_type == 0)
    {
        total_length = frame_length + PKT_HEADER_LEN;
    }
    else if (frame_type == 1)
    {
        total_length = frame_length + 10 + PKT_HEADER_LEN;
    }
    else if (frame_type == 2)
    {
        total_length = frame_length + 2 + PKT_HEADER_LEN;
    }

    ///打印报文, 每16字节一行，128字节一段
    for (i = 0; i < total_length; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        {
            printf(" ");
        }
        printf("%02x ", p8[i]);
    }
    printf("\n");
}

///@brief 接收报文的线程
///@param ptr 线程建立时的指针，这里指向线程的数据流号
void rx_thread(void *ptr)
{
    int sid, ret;
    void *ip;
    int len;   
    packet_t *pkt;
    struct ip *iph;
    uint8_t nl_proto;
    ///移植时，更换了类型；struct filter_full_rule与struct ratio_rule几近相同，
    ///建议作适当抽象后合并共同部分
    ///struct filter_full_rule r;
    struct ratio_rule r;
    struct pkt_header *pktinfo;
    int type;    
 
    sid = *((int *)ptr);
    
    ///循环，直到线程循环标志清零   
    while (glob_loop_flag_thread)
    {
        pkt = pag_get_packet(sid);
        
        ///--如果没有报文，则循环等待
        if (!pkt) 
        {
            usleep(500);
            continue;
        }
        
        ///--增加报文计数
        glob_pkts[sid][0]++;
        
        ///--获取网络层协议
        nl_proto = pag_get_nl_proto(pkt);
        if (nl_proto == 0xff)
        {
            printf("unknown proto\n");
            continue;
        }
        
        ///--获取网络层数据
        ip = pag_get_ip(pkt);
        
        ///--获取网络层数据长度
        len = pag_get_ip_length(pkt);
        
        glob_bytes[sid][0] += len;

        ///--打印报文（调试使用）
        if (glob_print_pkt_flag)
        {
            print_pkt(pkt, len);
        }

        ///--增加命中计数
        type = pag_getfiltertype(pkt);        
        if (type)
        {
            glob_hits[sid][0]++;
        }
        
        iph = (struct ip *)ip;
        ///--如果需要增加动态规则，则把当前报文的源ip构造规则，对规则表操作
        if (glob_rule_num > 0)
        {
            r.sip = htonl(iph->ip_src.s_addr);
            r.dip = 0;
            r.sport = 0;
            r.dport = 0;
            
            r.ruleid = 0;
            r.ruletype = sid + 1;
            
            r.optype = glob_rule_operate;
            r.iaction = FILTER_ACTION_PASS;
            r.send_action = OP_NONE;
            r.log_action = LOG_NONE;
            r.protocol = 0;
            
            strcpy(r.rule, "0:0:1:1");

            ret = pag_L4ratio_rule(sid, 1, &r);
            if (ret != -1)
            {
                glob_rule_num--;
            }
        }
    } //while (glob_loop_flag_thread)
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    struct timeval ts, te;
    double sec; 
    ulong all_pkts, all_bytes, all_hits;
    sigset_t set;
    pthread_t rx_tids[MAX_THREADS_NUM];
    int sids[MAX_THREADS_NUM];
    pthread_t print_tid;
    char rule_file[128];
    char ratio_file[128];
    int i;
    
    usage(argv[0]);
    if (argc > 8)
    {
        return -1;
    }
    
    ///读取参数
    rule_file[0] = '\0';
    ratio_file[0] = 0;
    glob_rule_num = 0;
    glob_rule_operate = FILTER_RULE_ADD;
    glob_print_pkt_flag = 0;

    while ((c = getopt(argc, argv, "f:r:t:s:ep?")) != -1)
    {
        switch (c)
        {
            case 'f':
                strcpy(rule_file, optarg);
                break;
            case 'r':
                glob_rule_num = atoi(optarg);
                break;
            case 't':
                glob_rule_operate = atoi(optarg);
                break;
            case 'p':
                glob_print_pkt_flag = 1;
                break;
            case 's':
                strcpy(ratio_file, optarg);
                break;
            case '?':
                usage(argv[0]);
		        return 0;
            default:
                break;
        }
    }
    
    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }
   
    ///如果指定了规则文件，则导入
    if (strlen(rule_file))
    {
        if (strlen(ratio_file))
        {
            pag_refreshratiorule(1, rule_file, ratio_file);
        }
        else
        {
            pag_refreshstaticrule(1, rule_file);
        }
    }
    
    ///创建线程
    glob_loop_flag_thread = 1;
    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        sids[i] = i;
        if (pthread_create(&rx_tids[i], NULL, (void *)rx_thread, 
            (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    if (pthread_create(&print_tid, NULL, (void *)print_thread,
            NULL) != 0)   
    {
        printf("### Error: Can not creat threads!\n");    
        pag_close();
        return -1;
    }

    ///等待结束信号 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    gettimeofday(&ts, NULL);   
     
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        usleep(500);  
    }
    
    gettimeofday(&te, NULL);  
    
    ///停止线程
    glob_loop_flag_thread = 0;
    
    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        pthread_join(rx_tids[i], NULL);
    }    
    pthread_join(print_tid, NULL);
      
    ///关闭设备
    pag_close();    
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Pkts recieved in threads: \n");
    all_pkts = 0;
    all_bytes = 0;
    all_hits = 0;
    for(i=0; i<MAX_THREADS_NUM; i++) 
    {
        printf("\t Thread[%d]:\t <%8lu> pkts, <%8lu M> bytes, <%8lu> hits\n", i,
            glob_pkts[i][0], glob_bytes[i][0] >> 20, glob_hits[i][0]);
        all_pkts += glob_pkts[i][0];
        all_bytes += glob_bytes[i][0];
        all_hits += glob_hits[i][0];
    }
    printf("Total : %lu pkts,  %lu bytes, %lu hits\n", 
        all_pkts, all_bytes, all_hits);
    printf("Speed : < %u > pps, < %u M > bps, < %u > hits ps\n", 
        (int)(all_pkts / sec), (int)((all_bytes >> 17) / sec), (int)(all_hits / sec));
    printf("\n");        

    return 0;
}
