////////////////////////////////////////
///@file test_stable.c
///
///@brief pag接口测试网卡稳定性，主线程加载规则，一个发包线程持续发送命中规则的报文。
///并根据所有收包线程的计数器判断正确性，如果正确，持续打印当前系统时间
///
/// 修改记录:
///
///@author lifengw@sugon.com
///@date 2012-06-08
///
///对不同的缓冲区调用不同的收包接口,包括pag_get_packet()和pag_get()。在老化中尽量测试到所有接口。
///
///@author lifengw@dawning.com.cn
///@date 2011-03-22
///
///增加对晶振频点的检测，如果频点不正确直接退出
///
///@author liuzhh@dawning.com.cn
///@date 2010-09-28
///
///增加对报文内容的判断，可配置随机的报文长度和是否使用过滤.
///对应rx线程处理时间增加，在使用-m参数时出现tx先于rx线程退出的问题，做了逻辑调整。
///修改默认规则类型为2，以保证不使用规则过滤时的判断有效。
///把结果输出文件操作从tx线程移动到main中，使文件是否存在表示是否开始测试
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-10
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <netinet/ip.h> //struct ip
#include <net/ethernet.h>

//本地路径下包含文件
#include "libpag.h"

///// 宏定义 ///////

#define MAX_THREADS_NUM 4  ///<线程数 
#define PRINT_ROUND_NUM 5  ///<每检查这些轮后打印一次信息
#define RULE_NUM 200000  ///<规则数
#define IP_START 0x0a000001  ///<规则的起始ip
#define IP_STEP 16  ///<规则ip的变化步长

#define DEF_PKT_LEN 40  ///<默认发送ip报文的长度，最小包
#define DEF_TEST_MINUTES 600000 ///<默认测试时间（分钟），无限时间
#define DEF_RULE_TYPE 2  ///<默认规则类型，2类。注意这里不能用1类，因为不使能规则时和命中1类规则一样，报文附加头规则类型为0

////////// 数据 /////////

static int glob_loop_flag_main;
static int glob_loop_flag_rx;
static int glob_loop_flag_tx;
static int glob_test_min;  ///<测试分钟数
static int glob_pkt_len;  ///<报文长度
static int glob_rule_type;  ///<规则类型

///计数器，每个线程的计数器内存空间分开，以避免cache抖动
static ulong glob_pkts[MAX_THREADS_NUM][16];
static ulong glob_errs[MAX_THREADS_NUM][16]; ///<内容出错的报文计数
static ulong glob_hits[MAX_THREADS_NUM][16];

///要发送的报文，数据内容在每次启动后会修改
static __u8 glob_pkt[1500] = 
        {
        0x45, 0x00, 0x00, 0x92, 0xc7, 0x6b, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01,
        0x02, 0x02, 0x02, 0x02 //ip head
        };

static int glob_result;
            
///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s [-m test_time_min] [-l ip_pkt_len] [-t rule_type]\n", cmd);
    printf("  -m test time in minutes, default = continously\n");
    printf("  -l ip pkt lenth, 0 or 40~1500, 0 is random, default = 40\n");
    printf("  -t rule type, 0~16, 0 means not checking rule filter, default = %d\n", DEF_RULE_TYPE);
    printf("\n");
}

///@brief 在全局报文结构中，构造要发送的报文，把ip头后的内容填充为随机数
void build_pkt()
{
    int i;
    struct iphdr *iph;
    
    iph = (struct iphdr *)glob_pkt;
    
    iph->tos = (__u8)rand();
    iph->id = (__u16)rand();
    iph->daddr = (__u32)rand();
    
    for (i = 20; i < 1500; i++)
    {
        glob_pkt[i] = (__u8)rand();
    }
}

///@brief 检查报文内容正确性，把收到的报文和发送的报文比较
///@param pkt 收到的报文
///@return 报文相同返回1，否则返回-1
int check_pkt_content(__u8 *pkt)
{
    int i, len;
    struct iphdr *iph_tx;
    struct iphdr *iph_rx;
    
    iph_tx = (struct iphdr *)glob_pkt;
    iph_rx = (struct iphdr *)pkt;
    
    
    if (iph_tx->tos != iph_rx->tos || iph_tx->id != iph_rx->id ||iph_tx->daddr != iph_rx->daddr)
    {
        return -1;
    }
    
    len = ntohs(iph_rx->tot_len);
    for (i = 20; i < len; i++)
    {
        if (glob_pkt[i] != pkt[i])
        {
            return -1;
        }
    }
    
    return 1;
}

///@brief 发送报文的线程
///@param ptr 线程建立时的指针，这里未使用
void tx_thread(void *ptr)
{
    int sid, type;
    void *pkt;   
    ulong all_tx, all_rx, all_hit, all_err;
    struct iphdr *iph;
    int i, j;
    time_t ts, te;
    int pkt_len;
    
    all_tx = 0;
    all_rx = 0;
    all_hit = 0;
    all_err = 0;
    
    ts = time(0);

    sid = 0;
    type = 0;
        
    ///循环，直到线程循环标志清零   
    glob_result = 0;
    while (glob_loop_flag_tx)
    {
        ///--打印测试起始时间和当前时间
        te = time(0);
        printf("\n******* Test time ********\n");
        printf("[From] %s", asctime(localtime(&ts)));
        printf("[To  ] %s", asctime(localtime(&te)));

        ///--循环指定轮数，每轮测试对所有规则循环发送命中报文
        for (i = 0; i < PRINT_ROUND_NUM; i++) 
        {
            for (j = 0; j < RULE_NUM; j++)
            {
                ///----申请发包缓冲区
                do
                {
                    pkt = pag_getsendbuf(sid);
                    if (pkt == NULL) 
                    {
                        if (glob_loop_flag_tx == 0)
                        {
                            return;
                        }
                        usleep(10);
                    }
                }
                while (pkt == NULL);
                
                ///----修改报文ip,从固定的ip开始按固定步长增加     
                if (glob_pkt_len == 0)
                {
                    pkt_len = (rand() % 1460) + 40;
                }
                else
                {
                    pkt_len = glob_pkt_len;
                }
              
                memcpy(pkt, glob_pkt, pkt_len);   
                iph = (struct iphdr *)pkt;
                iph->saddr = htonl(IP_START + (j * IP_STEP));
                iph->tot_len = htons(pkt_len);
                
                ///----发送报文
                pag_send(type, sid, pkt_len);
                all_tx++;
            } //for (j = 0; j < RULE_NUM; j++)
        } //for (i = 0; i < PRINT_ROUND_NUM; i++) 
        
        ///--检查收包和发包数是否一致
        printf("\nchecking packets count of tx and rx ...... ");
        
        ///--统计收包总数时需要等待所有收包线程处理完毕
        for (i = 0; i < 3000; i++)        
        {
            all_rx = 0;
            for (j = 0; j < MAX_THREADS_NUM; j++) 
            {
                all_rx += glob_pkts[j][0];
            }
            if (all_rx != all_tx)
            {
                usleep(1000);
            }
            else
            {
                break;
            }
        }

        if (all_rx == all_tx)
        {
            printf("OK\n");
        }
        else
        {
            glob_result = 1;
            printf("FAILED\n");
            
            glob_loop_flag_main = 0;
            glob_loop_flag_tx = 0;
        }
        printf("rx pkts = %lu, tx pkts = %lu\n", all_rx, all_tx);
        
        all_err = 0;
        for (i = 0; i < MAX_THREADS_NUM; i++) 
        {
            all_err += glob_errs[i][0];
        }
        if (all_err)
        {
            glob_result = 1;
            printf("error pkts = %llu\n", all_err);
            
            glob_loop_flag_main = 0;
            glob_loop_flag_tx = 0;
        }
        
        ///--如果使能规则，则检查收包和命中计数是否一致
        if (glob_rule_type)
        {
            printf("\nchecking packets count of rx and hit ...... ");
            all_hit = 0;
            for (i = 0; i < MAX_THREADS_NUM; i++) 
            {
                all_hit += glob_hits[i][0];
            }
            
            if (all_rx == all_hit)
            {
                printf("OK\n");
            }
            else
            {
                glob_result = 1;
                printf("FAILED\n");
                
                glob_loop_flag_main = 0;
                glob_loop_flag_tx = 0;
            }
            printf("rx pkts = %llu, hit = %llu\n", all_rx, all_hit);
        }
        
    } //while (glob_loop_flag_thread)     
}


///@brief 接收报文的线程
///@param ptr 线程建立时的指针，这里指向线程的数据流号
void rx_thread(void *ptr)
{
    int sid;
    packet_t *packet;
    void *pkt;
    uint8_t nl_proto;     
    int type;
    
    sid = *((int *)ptr);
        
    ///循环，直到线程循环标志清零   
    while (glob_loop_flag_rx)
    {
        if (!(sid % (MAX_THREADS_NUM / 2)))
        {
            packet = pag_get_packet(sid);
        
            ///--如果没有报文，则循环等待
            if (!packet) 
            {
                usleep(500);
                continue;
            }

            ///--获取网络层协议
            nl_proto = pag_get_nl_proto(packet);
            if (nl_proto == 0xff)
            {
                printf("unknown proto\n");
                continue;
            }
        
            ///--获取ip报文头
            pkt = pag_get_ip(packet);
        }
        else
        {
            ///--获取ip报文头
            pkt = pag_get(sid);
            
            ///--如果没有报文，则循环等待
            if (!pkt) 
            {
                usleep(500);
                continue;
            }   
            
            ///--偏移到附加头位置
            packet = (packet_t *)((uint8_t *)pkt - sizeof(struct ether_header) - 50);
        }
        
        ///--增加报文计数
        if (check_pkt_content((__u8 *)pkt) == 1)
        {
            glob_pkts[sid][0]++;
        }
        else
        {
            glob_errs[sid][0]++;
        }
        
        ///--如果使能规则，则增加命中计数
        if (glob_rule_type)
        {
            type = pag_getfiltertype(packet);        
            if (type == glob_rule_type)
            {
                glob_hits[sid][0]++;
            }   
        }      
    }
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
    glob_loop_flag_tx = 0;
}

///@brief 向网卡中添加规则
void add_test_rules()
{
    int i;
    struct filter_full_rule rules[RULE_NUM];
    struct filter_full_rule *r;
    
    if (glob_rule_type == 0)
    {
        return;
    }
    
    ///循环构造规则表，规则ip从固定的ip开始,按固定步长增加
    for (i = 0; i < RULE_NUM; i++)
    {
        r = &rules[i];
        r->sip = IP_START + (i * IP_STEP);
        r->dip = 0;
        r->sport = 0;
        r->dport = 0;
        
        r->ruleid = 0;
        r->ruletype = glob_rule_type;
        
        r->optype = FILTER_RULE_ADD;
        r->iaction = FILTER_ACTION_PASS;
        r->send_action = OP_NONE;
        r->log_action = LOG_NONE;
        r->protocol = 0;
    }
    
    ///向网卡添加规则表
    pag_L4rule(0, RULE_NUM, rules);
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    time_t ts, te;
    int min; 
    sigset_t set;
    pthread_t rx_tids[MAX_THREADS_NUM];
    int sids[MAX_THREADS_NUM];
    pthread_t tx_tid;
    int i;
    
    ///结果文件空，表示测试开始
    system("echo >> /tmp/netfirm.result");
    
    ///检测频点
    system("./../tools/nic_dumphw_mem 7a00 > /dev/null 2>&1");
    if (system("./../tools/nic_dumphw_mem 7a00 | grep value=ff > /dev/null 2>&1") == 0)
    {
        fprintf(stderr, "##Error:crystal oscillator frequency error.\n");
        return -1;
    }
    
    usage(argv[0]);       
    if (argc > 7)
    {
        return -1;
    }
    
    ///读取参数
    glob_rule_type = DEF_RULE_TYPE;
    glob_test_min = DEF_TEST_MINUTES;
    glob_pkt_len = DEF_PKT_LEN;
    while ((c = getopt(argc, argv, "m:l:t:?")) != -1) 
    {
        switch (c)
        {
            case 'm':
                glob_test_min = atoi(optarg);
                break;
            case 'l':
                glob_pkt_len = atoi(optarg);
                if (glob_pkt_len == 0 || (glob_pkt_len >= 40 && glob_pkt_len <= 1500))
                {
                    break;
                }
                else
                {
                    printf("\nError: pkt lenth invalid!\n");
                    usage(argv[0]);
                    return -1;
                }
            case 't':
                glob_rule_type = atoi(optarg);
                if (glob_rule_type >= 0 && glob_rule_type <= 16)
                {
                    break;
                }
                else
                {
                    printf("\nError: rule type invalid!\n");
                    usage(argv[0]);
                    return -1;
                }
            case '?':
            default:
                usage(argv[0]);
                return 0;
                break;
        }
    }

    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }
    
    ///添加规则
    add_test_rules();
    
    ///构造要发送的报文
    build_pkt();
    
    ///创建线程
    glob_loop_flag_rx = 1;
    glob_loop_flag_tx = 1;
    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        sids[i] = i;
        if (pthread_create(&rx_tids[i], NULL, (void *)rx_thread, 
            (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    if (pthread_create(&tx_tid, NULL, (void *)tx_thread,
            NULL) != 0)   
    {
        printf("### Error: Can not creat threads!\n");    
        pag_close();
        return -1;
    }

    ///等待结束信号，或着收到期望的包数 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    ts = time(0);
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        sleep(1);  
        
        te = time(0);
        min = (int)((te - ts) / 60);
        if (glob_test_min != 0 && min >= glob_test_min)
        {
            break;
        }
    }
    
    ///停止线程，先停发包，确认发包停止后，再停收包
    glob_loop_flag_tx = 0;    
    pthread_join(tx_tid, NULL);
    
    glob_loop_flag_rx = 0;
    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        pthread_join(rx_tids[i], NULL);
    }    
      
    ///关闭设备
    pag_close();    
    
    ///输出到结果文件
    if (glob_result)
    {
        system("echo failed >> /tmp/netfirm.result");
        return 1;
    }
    else
    {
        system("echo passed >> /tmp/netfirm.result");
        return 0;   
    }
}

