////////////////////////////////////////
///@file tx_pcap_sample.c
///
///@brief 发送pcap文件，模仿tcp_replay，
///程序使用pag.conf文件获得设备名，使用eth层发包函数，可以指定发包端口和循环发送次数
///
///修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2010-03-13
///
///在pos网卡发包程序tx_real基础上修改，建立文档
///
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //sleep
#include <pthread.h>  //pthread_create
#include <linux/types.h>  //ulong
#include <string.h>  //memcpy

#include <pcap.h>
#include <libpag.h>


///// 宏定义 ///////

#define TX_SID 0  ///<默认使用0号发包缓冲区

//数据链路层的类型
#ifndef DLT_EN10MB
#define DLT_EN10MB 1  ///<带ethernet头的以太包
#endif
#ifndef DLT_RAW
#define DLT_RAW 14   ///<原始ip包
#endif

////////// 数据 /////////

ulong glob_pkts;
ulong glob_bytes;

int glob_mac_port;  ///<发包使用的网口
long start_sn;  ///<文件中开始发的第一个包的编号
long total_pkt;  ///<发包个数
int glob_pkt_offset; ///<报文在发包缓冲区内的偏移，ip为14，ethernet为0
    
///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s <loop times> <mac port> <start pakcet num> <total send packet num> <pcap file> ", cmd);
    printf("\n");
}


///@brief 发送报文的回调函数,从指定端口发送ethernet包，累加包数，字节数
void callback_tx_pkt(u_char *argument, const struct pcap_pkthdr *packet_header, 
    const u_char* packet_content)
{
    void *pkt_buf;
    __u8 *p8;
    int len;

    if (start_sn != 0)
    {
        start_sn--;
        return;
    }
    
    ///--申请发包缓冲区，如果没有，则循环等待
    pkt_buf = pag_getsendbuf(TX_SID);
    if (!pkt_buf) 
    {
        usleep(500);
        return;
    }
    
    len = packet_header->len;
    
    ///--把报文拷贝到发送缓冲区，如果是原始IP包，则构造一个mac头
    memcpy((void *)pkt_buf + glob_pkt_offset, (void *)packet_content, len);
    
    ///--发送报文
    if (pag_send_eth(glob_mac_port, TX_SID, len) == -1)
    {
        return;
    }

    ///--累计发包数和字节数
    glob_pkts++;
    glob_bytes += len;
    total_pkt--;
    
    if (total_pkt == 0)
    {
        exit(0);
    }
    
    //printf("glob pkts = %lu,  glob bytes = %lu\n", glob_pkts, glob_bytes);
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    pcap_t *pcap_handle;
    char error[PCAP_ERRBUF_SIZE];
    char *file;
    int c, loop_times, mac_port;
    struct timeval ts, te;
    double sec; 
    int i;
    
    ///判断参数，出错时打印使用信息
    if (argc != 6)
    {
        usage(argv[0]);
        return -1;
    }
    
    ///读取参数
    loop_times = atoi(argv[1]);
    glob_mac_port = atoi(argv[2]);
    start_sn = atoi(argv[3]);
    total_pkt = atoi(argv[4]);
    file = argv[5];


    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }

    glob_pkts = 0;
    glob_bytes = 0;
    
    gettimeofday(&ts, NULL); 
    
    ///循环发送文件    
    printf("\n Reading pkts to send ......\n");
    for (i = 0; i < loop_times; i++) 
    {
        ///--打开pcap文件句柄
        pcap_handle = pcap_open_offline(file, error);
        if (pcap_handle == NULL)
        {
            printf("open file error: %s\n", error);
            break;
        }
        
        ///--读取数据链路层封装格式
        if (pcap_datalink(pcap_handle) == DLT_EN10MB)
        {
            glob_pkt_offset = -14;
        }
        else
        {
            glob_pkt_offset = 0;
        }
        
        ///--从文件中循环收包，用回调函数发送每个包
        if (pcap_loop(pcap_handle, -1, callback_tx_pkt, NULL) < 0)
        {
            break;
        }
        
        ///--关闭文件句柄
        pcap_close(pcap_handle);
 
    }
    
    gettimeofday(&te, NULL);  
    
    ///关闭设备
    pag_close();   
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Pkts sent summary: \n");
    printf("Total : %lu pkts,  %lu bytes\n", glob_pkts, glob_bytes);
    printf("Speed : < %u > pps, < %u M > bps\n", 
        (int)(glob_pkts / sec), (int)((glob_bytes >> 17) / sec));
    printf("\n");        
    return 0;
}
