////////////////////////////////////////
///@file tx_sample.c
///
///@brief pag接口发包示例代码，可以指定发包的长度，类型，源目ip
///
///修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2010-09-29
///
///整理命令行参数，使用更方便
///默认线程数和最大线程数的宏定义分开定义
///
///
///@author nisp@dawning.com.cn
///@date 2010-03-19
///
///增加-h的命令行提示
///修正发包数量不能按照-n指定的数目进行发送的bug
///增加-m选项，可支持2.0设计的32线程发包，默认4线程发包
///
///@author liuzhh@dawning.com.cn
///@date 2010-02-23
///
///增加发送随机协议和ip功能
///
///@author liuc@dawning.com.cn
///@date 2010-01-10
///
///增加发送随机包长功能
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-10
///
///建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //getopt
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <linux/types.h>
#include <net/ethernet.h>
#include <netinet/ip.h>  //struct ip
#include <netinet/in.h>  //IPPROTO_TCP
#include <arpa/inet.h>  //inet_addr();

#include <libpag.h>

///// 宏定义 ///////

#define DEF_THREADS_NUM 1   ///<默认线程数
#define MAX_THREADS_NUM 32  ///<系统支持的最大线程数 

#define MIN_TX_PKT_LEN  28  ///<可发送的最小报文长度
#define MAX_TX_PKT_LEN  3500  ///<可发送的最大报文长度
#define MAX_TX_MAC_IN_MAC_PKT_LEN 4096 ///<可发送的最大MAC IN MAC报文长度


////////// 数据 /////////

static int glob_threads_num;
static int glob_loop_flag_main;
static int glob_loop_flag_thread;
static int glob_pkts_num;
static int glob_pkts_num_thread[MAX_THREADS_NUM];
static int glob_inc_len;
static int glob_print_pkt_flag;
static int glob_pkt_low_len;
static int glob_pkt_up_len;
static int glob_pkt_proto;
static __u32 glob_sip;
static __u32 glob_dip;

///计数器，每个线程的计数器内存空间分开，以避免cache抖动
static ulong glob_pkts[MAX_THREADS_NUM][16];
static ulong glob_bytes[MAX_THREADS_NUM][16];


///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s ", cmd);
    printf("[-n pkt_num] [-m thread_num] [-l low_len] [-u up_len] [-i increase_step] \n");
    printf("[-t pkt_type(proto)] [-s src_ip] [-d dst_ip] [-p] [-?]\n");
    printf("  -n pkts num to tx, default = continously\n");
    printf("  -m send thread number, default = %d\n", DEF_THREADS_NUM);
    printf("  -l min pkt lenth, default = %d\n", MIN_TX_PKT_LEN);
    printf("  -u max pkt lenth, default = %d\n", MAX_TX_PKT_LEN);
    printf("  -i pkt length increase, 0 = random, default = 0 \n");
    printf("  -t pkt protocol and type, 0 = random(mac), 17 = udp(log), 6 = tcp(rst), default = 0\n");
    printf("  -s src ip of pkt, 0 = random, default = 0\n");
    printf("  -d dst ip of pkt, 0 = random, default = 0\n");
    printf("  -p print pkts and headers\n");
    printf("  -? print usage message\n");
    printf("\n");
}


///@brief 打印发包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_thread(void *ptr)
{
    ulong old_pkts[MAX_THREADS_NUM];
    ulong old_bytes[MAX_THREADS_NUM];
    int i;

    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        old_pkts[i] = 0;
        old_bytes[i] = 0;
    }
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        printf("\nCurrent speed-----\n");
        for (i = 0; i < glob_threads_num; i++) 
        {
            printf("\t Thread[%d]:\t %8lu pps, %8lu M bps\n", i, 
                glob_pkts[i][0] - old_pkts[i], 
                (glob_bytes[i][0] - old_bytes[i]) >> 17);
            old_pkts[i] = glob_pkts[i][0];
            old_bytes[i] = glob_bytes[i][0];
        }
    }
}

///@brief 打印报文（检查收包内容）
///@param pkt 指向报文的指针
///@param len 报文长度
static void print_pkt(void *pkt, int len)
{
    int i;
    __u8 *p8;
    static int pkt_print;
 
    ///找到报文附加头, 打印长度包括附加头
    pkt_print++;
    p8 = (__u8 *)pkt;

    ///打印报文, 每16字节一行，128字节一段
    printf("\n====  pkt:%d  ====\n", pkt_print);
    for (i = 0; i < len; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        printf(" ");
        printf("%02x ", p8[i]);
    }
    printf("\n");
}

///@brief 构造报文
///@param pkt 指向报文的指针
///@param len 报文长度
///@param proto 报文类型，17为日志包（udp）,6为封堵包（tcp）
inline void build_pkt(void *pkt, int len, int proto, __u32 sip, __u32 dip)
{
    struct iphdr *iph;
    
    ///找到报文ip头
    iph = (struct iphdr *)pkt;
    
    ///修改各字段
    iph->tot_len = htons(len);
    iph->protocol = proto;
    iph->saddr = sip;
    iph->daddr = dip;
}

///@brief 发送报文的线程
///@param ptr 线程建立时的指针，这里指向线程的数据流号
void tx_thread(void *ptr)
{
    int i, r, sid, len, proto, type;
    __u32 sip, dip;
    void *pkt;
    unsigned char *p;
    int temp = 0;
    int range;
    __u8 p8[MAX_TX_PKT_LEN] = 
        {
        0x45, 0x00, 0x00, 0x92, 0xc7, 0x6b, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01,
        0x02, 0x02, 0x02, 0x02, //ip head
        0x00, 0x50, 0x07, 0xbb, 0x78, 0x28, 0x73, 0x3e, 
        0x43, 0x23, 0x6a, 0x7c, 0x50, 0x12, 0x16, 0x58, 
        0x76, 0xd4, 0x00, 0x00//tcp head
        };
    
#ifndef _CONFIG_C27    
    ///把报文的应用数据设置为递增字节
    for (i = 0 ; i < MAX_TX_PKT_LEN - 40; i++)
    {
        p8[40 + i] = (__u8)i;
    }
    sid = *((int *)ptr);
    
    srand(time(0));   

    ///循环，直到线程循环标志清零
    while (glob_loop_flag_thread)
    {
        if (glob_pkts_num_thread[sid] != -1 && glob_pkts[sid][0] == glob_pkts_num_thread[sid])
        {
            break;
        }
        
        ///--如果没有缓冲区，则循环等待
        while (!(pkt = pag_getsendbuf(sid)))
        {
            usleep(500);
        }
        
        r = rand();
        range = glob_pkt_up_len - glob_pkt_low_len + 1;
        if (glob_inc_len)
        {
            len = temp % range + glob_pkt_low_len;
            temp += glob_inc_len;
        }
        else
        {
            len = r % range + glob_pkt_low_len;
        }
        
        ///--把原始报文拷贝到发送缓冲区，根据参数对报文修改报文内容
        memcpy(pkt, p8, len);
        
        ///--如果指定了报文协议，则协议决定报文在sendlog.conf中的封堵和日志类型
        ///--否则，随机确定协议，类型为mac帧
        if (glob_pkt_proto)
        {
            proto = glob_pkt_proto;
            
            if (proto == IPPROTO_UDP)
            {
                type = 0;
            }
            else
            {
                type = 1;
            }
        }
        else
        {
            if (r & 1)
            {
                proto = IPPROTO_UDP;
                type = 0;
            }
            else
            {
                proto = IPPROTO_TCP;
                type = 1;
            }
        }
        
        if (glob_sip)
        {
            sip = glob_sip;
        }
        else
        {
            sip = (__u32)r;
        } 
        
        if (glob_dip)
        {
            dip = glob_dip;
        }
        else
        {
            dip = (__u32)(~r);
        } 
        
        build_pkt(pkt, len, proto, sip, dip);

        ///--发送报文
        if (pag_send(type, sid, len) == -1)
        {
            continue;
        }
       
        if (glob_print_pkt_flag)
        {
            print_pkt(pkt, len);
        }
        
        ///--增加报文计数
        glob_pkts[sid][0]++;
        glob_bytes[sid][0] += len;

    }
#else
    __u8 mim_p8[MAX_TX_MAC_IN_MAC_PKT_LEN] = 
        {
        0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x22, 0x22,
        0x22, 0x22, 0x22, 0x22, 0x88, 0xa8, 0xaa, 0xaa,
        0xaa, 0xaa, 0xaa, 0xaa, 0xbb, 0xbb, 0xbb, 0xbb,
        0xbb, 0xbb, 0x08, 0x00, //MAC IN MAC head
        0x45, 0x00, 0x00, 0x92, 0xc7, 0x6b, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01,
        0x02, 0x02, 0x02, 0x02, //ip head
        0x00, 0x50, 0x07, 0xbb, 0x78, 0x28, 0x73, 0x3e, 
        0x43, 0x23, 0x6a, 0x7c, 0x50, 0x12, 0x16, 0x58, 
        0x76, 0xd4, 0x00, 0x00//tcp head
        };
        
    ///把报文的应用数据设置为递增字节
    for (i = 0 ; i < MAX_TX_MAC_IN_MAC_PKT_LEN - (ETHER_HDR_LEN << 1) - 40; i++)
    {
        mim_p8[(ETHER_HDR_LEN << 1) + 40 + i] = (__u8)i;
    }
    sid = *((int *)ptr);
    
    srand(time(0));   

    ///循环，直到线程循环标志清零
    while (glob_loop_flag_thread)
    {
        if (glob_pkts_num_thread[sid] != -1 && glob_pkts[sid][0] == glob_pkts_num_thread[sid])
        {
            break;
        }
        
        ///--如果没有缓冲区，则循环等待
        while (!(pkt = pag_getsendbuf(sid)))
        {
            usleep(500);
        }
        
        r = rand();
        range = glob_pkt_up_len - glob_pkt_low_len + 1;
        if (glob_inc_len)
        {
            len = temp % range + glob_pkt_low_len;
            temp += glob_inc_len;
        }
        else
        {
            len = r % range + glob_pkt_low_len;
        }
        
        ///--把原始报文拷贝到发送缓冲区，根据参数对报文修改报文内容
        memcpy(pkt, mim_p8, len);
        
        if (glob_sip)
        {
            sip = glob_sip;
        }
        else
        {
            sip = (__u32)r;
        } 
        
        if (glob_dip)
        {
            dip = glob_dip;
        }
        else
        {
            dip = (__u32)(~r);
        } 
        
        proto = IPPROTO_TCP;
        
        build_pkt(pkt + (ETHER_HDR_LEN << 1), len - (ETHER_HDR_LEN << 1), proto, sip, dip);

        ///--发送报文
        if (pag_send_eth(sid, pkt, len) == -1)
        {
            continue;
        }
       
        if (glob_print_pkt_flag)
        {
            print_pkt(pkt, len);
        }
        
        ///--增加报文计数
        glob_pkts[sid][0]++;
        glob_bytes[sid][0] += len - (ETHER_HDR_LEN << 1);
    }
#endif
}


///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c;
    struct timeval ts, te;
    double sec; 
    ulong all_pkts, all_bytes;
    sigset_t set;
    pthread_t tx_tids[MAX_THREADS_NUM];
    int sids[MAX_THREADS_NUM];
    pthread_t print_tid;
    int grain, remain;
    int i;
    
    ///打印使用信息
    usage(argv[0]);
    
    ///读取参数
    glob_pkts_num = -1;
    glob_inc_len = 0;
    glob_pkt_low_len = MIN_TX_PKT_LEN;
    glob_pkt_up_len = MAX_TX_PKT_LEN;
    glob_pkt_proto = 0;
    glob_sip = 0;
    glob_dip = 0; 
    glob_threads_num = DEF_THREADS_NUM;
    glob_print_pkt_flag = 0;
    for (i = 0; i < MAX_THREADS_NUM; i++)
    {
        glob_pkts_num_thread[i] = 0;
    }
   
    while ((c = getopt(argc, argv, "n:m:l:u:i:t:s:d:p?")) != -1) 
    {
        switch (c)
        {
            case 'n':
                glob_pkts_num = atoi(optarg);
                break;
            case 'm':
                glob_threads_num = atoi(optarg);
                if (glob_threads_num > MAX_THREADS_NUM)
                {
                    printf("\nError: threads num invalid\n");
                    return -1;
                }
                break;
            case 'l':
                glob_pkt_low_len = atoi(optarg);
                if (glob_pkt_low_len < MIN_TX_PKT_LEN || glob_pkt_low_len > MAX_TX_PKT_LEN)
                {
                    printf("\nError: min pkt len invalid\n");
                    return -1;
                }
                break;
            case 'u':
                glob_pkt_up_len = atoi(optarg);
                if (glob_pkt_up_len < MIN_TX_PKT_LEN || glob_pkt_up_len > MAX_TX_PKT_LEN)
                {
                    printf("\nError: max pkt len invalid\n");
                    return -1;
                }
                break;
            case 'i':
                glob_inc_len = atoi(optarg);
                break;
            case 't':
                glob_pkt_proto = atoi(optarg);
                break;
            case 's':
                glob_sip = inet_addr(optarg);
                break;
            case 'd':
                glob_dip = inet_addr(optarg);
                break;
            case 'p':
                glob_print_pkt_flag = 1;
                break;
            case '?':
            default:
                usage(argv[0]);
                return 1;
        }
    }

    if (glob_pkt_up_len < glob_pkt_low_len)
    {
        printf("Warning:send len rang[%d, %d]\n", glob_pkt_low_len, glob_pkt_up_len);
        return -1;
    }
    
    ///将-n指定的发包数均分到各个线程发送
    if (glob_pkts_num != -1)
    {
        grain = glob_pkts_num / glob_threads_num;
        remain = glob_pkts_num % glob_threads_num;
    }
    
    for (i = 0; i < glob_threads_num; i++)
    {
        if (glob_pkts_num == -1)
        {
            glob_pkts_num_thread[i] = -1;
        }
        else
        {
            glob_pkts_num_thread[i] = grain;
            if (i == 0)
            {
                glob_pkts_num_thread[i] += remain;
            }
        }
    }
    
    ///打开设备
    if (pag_open() < 0)
    {
        printf("### Error : can not open device! ###\n");
        return -1;
    }
    
    gettimeofday(&ts, NULL);   
    
    ///创建线程
    glob_loop_flag_thread = 1;
    for (i = 0; i < glob_threads_num; i++)
    {
        sids[i] = i;
        if (pthread_create(&tx_tids[i], NULL, (void *)tx_thread, 
            (void *)&sids[i]) != 0)   
        {
            printf("### Error: Can not creat threads!\n");    
            pag_close();
            return -1;
        }
    }
    
    ///打印报文速度
    if (pthread_create(&print_tid, NULL, (void *)print_thread,
            NULL) != 0)   
    {
        printf("### Error: Can not creat threads!\n");    
        pag_close();
        return -1;
    }

    ///等待结束信号，或着发出期望的包数 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        usleep(10);  
        
        all_pkts = 0;
        all_bytes = 0;
        for (i = 0; i < glob_threads_num; i++) 
        {
            all_pkts += glob_pkts[i][0];
        }
        
        if (glob_pkts_num != 0 && glob_pkts_num <= all_pkts)
        {
            break;
        }
    }
    
    ///停止线程
    glob_loop_flag_thread = 0;
            
    gettimeofday(&te, NULL);  
    
    for (i = 0; i < glob_threads_num; i++)
    {
        pthread_join(tx_tids[i], NULL);
    }
    pthread_join(print_tid, NULL);
      
    ///关闭设备
    pag_close();   
    
    ///打印总结信息
    sec = (float)(te.tv_sec - ts.tv_sec) + ((float)(te.tv_usec - ts.tv_usec)) 
        / 1000000.f;   
    printf("\n");        
    printf("Pkts sent in threads: \n");
    all_pkts = 0;
    all_bytes = 0;
    for (i = 0; i < glob_threads_num; i++) 
    {
        printf("\t Thread[%d]:\t <%8lu> pkts, <%8lu M> bytes\n", i,
            glob_pkts[i][0], glob_bytes[i][0] >> 20);
        all_pkts += glob_pkts[i][0];
        all_bytes += glob_bytes[i][0];
    }
    printf("Total : %lu pkts,  %lu bytes\n", all_pkts, all_bytes);
    printf("Speed : < %u > pps, < %u M > bps\n", 
        (int)(all_pkts / sec), (int)((all_bytes >> 17) / sec));
    printf("\n");        
    return 0;
}
