////////////////////////////////////////
///@file pcap_mult_nic_poll_sample.c
///
///@brief pcap接口收包多网卡示例代码，主线程从多个网卡轮询收包
///
/// 修改记录:
///
///@author liuzhh@dawning.com.cn
///@date 2011-02-26
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //sleep
#include <pthread.h>  //pthread_create
#include <signal.h>
#include <string.h>
#include <linux/types.h>  //ulong
#include <netinet/ip.h>  //iphdr
#include <netinet/tcp.h>  //tcphdr

#include <pcap.h>

///// 宏定义 ///////


#define MAX_NIC_NUM 16  ///<收包网卡最大个数 

#define SPEC_PORT 80  ///<需要检查的特定端口

////////// 数据 /////////

int glob_loop_flag_main;
int glob_loop_flag_thread;

char glob_dev_name[MAX_NIC_NUM][16];
int glob_nic_num;  ///<实际使用的网卡数
ulong glob_pkts_nic[MAX_NIC_NUM]; ///<每个网卡（线程）收包数
ulong glob_bytes_nic[MAX_NIC_NUM];
ulong glob_hits_nic[MAX_NIC_NUM];
ulong glob_pkts;  ///<所有网卡（线程）总收包数
ulong glob_bytes;
ulong glob_hits;

///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd) 
{
    printf("Usage: %s dev0 [dev1] [...]\n", cmd);
    printf("   dev?: dev_name[:stream_id]\n");
    printf("   dev_name: device name shown in ifconfig\n");
    printf("   stream_id: must < stream num in $dev_name.conf\n");
    printf("                default = get pkts from all streams\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_speed_thread(void *ptr)
{
    ulong old_pkts;
    ulong old_bytes;
    ulong old_hits;
    int i;

    old_pkts = 0;
    old_bytes = 0;
    old_hits = 0;
    
    while (glob_loop_flag_thread)
    {
        sleep(1); 
        
        glob_bytes = 0;
        glob_pkts = 0;
        glob_hits = 0;
        for (i = 0; i < glob_nic_num; i++)
        {
            printf("NIC %d: pkts = %10llu, bytes = %10llu, hits = %10llu\n", i, 
            glob_pkts_nic[i], glob_bytes_nic[i], glob_hits_nic[i]);
            
			glob_pkts += glob_pkts_nic[i];
			glob_bytes += glob_bytes_nic[i];
			glob_hits += glob_hits_nic[i];
		}
		
        printf("\nCurrent speed: %8lu pps, %8lu M bps, %8lu hits ps\n", glob_pkts - old_pkts, 
            (glob_bytes - old_bytes) >> 17,
            glob_hits - old_hits);
            
            old_pkts = glob_pkts;
            old_bytes = glob_bytes;
            old_hits = glob_hits;
    }
}

///@brief 处理报文的回调函数,累加收包数，字节数和命中指定端口的包数
void callbak_process_pkt(u_char *argument, const struct pcap_pkthdr *packet_header, const u_char* packet_content)
{
    struct iphdr *iph;
    struct tcphdr *tcph;
    int tid;
    
    tid = *((int *)argument);
    
    glob_pkts_nic[tid]++;
    
    iph = (struct iphdr *)(packet_content + 14);    
    glob_bytes_nic[tid] += ntohs(iph->tot_len);
    
    if(iph->protocol != 0x6)
    {
        return;
    }
    
    tcph = (struct tcphdr *)(packet_content + 34);
    if(ntohs(tcph->source) == SPEC_PORT || ntohs(tcph->dest) == SPEC_PORT) 
    {
        glob_hits_nic[tid]++;
    }
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_main = 0;
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    pthread_t print_tid;
    sigset_t set;
    pcap_t *pcap_handle[MAX_NIC_NUM];
    char error_content[PCAP_ERRBUF_SIZE];
    int i;

    ///读取参数
    if (argc < 2)
    {
        usage(argv[0]);
        return -1;
    }

    glob_nic_num = argc - 1;
    	
    ///创建打印速度的线程
    glob_loop_flag_thread = 1;
    if (pthread_create(&print_tid, NULL, (void *)print_speed_thread, NULL) != 0)   
    {
        printf("### Error: Can not creat threads!\n");    
        return -1;
    }

    ///循环打开所有设备
    for (i = 0; i < glob_nic_num; i++)
    {
		strcpy(glob_dev_name[i], argv[i + 1]);
		if ((pcap_handle[i] = pcap_open_live(glob_dev_name[i], BUFSIZ, 1, 0, error_content)) == NULL)
        {
            printf("Erorr: pcap open device [%s] failed!\n", glob_dev_name[i]);
            return -1;
        }
    }
    
    ///等待结束信号 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    ///在收到中断信号前循环，依次从每个网卡收包，用回调函数处理每个包
    glob_loop_flag_main = 1;
    while (glob_loop_flag_main) 
    {
        for (i = 0; i < glob_nic_num && glob_loop_flag_main; i++)
        {
            pcap_read(pcap_handle[i], 1, callbak_process_pkt, (void *)&i);
        }
    }

    ///停止打印速度的线程
    glob_loop_flag_thread = 0;
    pthread_join(print_tid, NULL);

    ///关闭设备
    for (i = 0; i < glob_nic_num; i++)
    {
        pcap_close(pcap_handle[i]);
    }            

    ///打印结果
    printf("\n Total : %lu pkts, %lu bytes, %lu hits\n", 
        glob_pkts, glob_bytes, glob_hits);

    return 0;
}
