////////////////////////////////////////
///@file pcap_sample.c
///
///@brief pcap接口收包示例代码，收包线程的个数需要小于或等于配置文件中的分流数，
///如果不指定收包线程数，则用主应用收包，而不启动线程
///
/// 修改记录:
///
///@author liucan@dawning.com.cn
///@date 2010-04-02
/// 修改内容：添加调试所用的打印数据包内容函数，添加-p命令参数！
///
///@date 2010-03-25
/// 修改内容：添加收指定包数的功能，参数格式使用有所改变！
///
///@author lifengw@dawning.com.cn
///@date 2009-12-25
///
///@author liuzhh@dawning.com.cn
///@date 2009-11-13
///
/// 修改内容：建立文档
////////////////////////////////////////

//系统路径下头文件
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  //sleep
#include <pthread.h>  //pthread_create
#include <linux/types.h>  //ulong
#include <netinet/ip.h>  //iphdr
#include <netinet/tcp.h>  //tcphdr
#include <signal.h>

#include <pcap.h>

///// 宏定义 ///////


#define MAX_THREADS_NUM 1  ///<收包线程最大个数

#define SPEC_PORT 80  ///<需要检查的特定端口

//数据链路层的类型
#ifndef DLT_EN10MB
#define DLT_EN10MB 1  ///< 带ethernet头的以太包
#endif
#ifndef DLT_RAW
#define DLT_RAW 14   ///< 原始ip包
#endif

////////// 数据 /////////

int glob_loop_flag_thread;
pcap_t *glob_pcap_handle;
int glob_print_pkt_flag = 0;
int glob_print_port_flag = 0;
static glob_compare_content_flag;

ulong glob_pkts;
ulong glob_bytes;
ulong glob_tcp_hits;
ulong glob_udp_hits;
ulong glob_filter_flg = 0;

///////   函数列表  ///////////////

///@brief 输出程序帮助信息
///@param cmd 程序命令(输入)
void usage(char* cmd)
{
    printf("Usage: %s [-i dev_name:stream_id] [-n pkt_num] [-p] [-o]\n", cmd);
    printf("  -i dev_name:stream_id device name shown in ifconfig\n");
    printf("     stream_id: must < stream num in $dev_name.conf\n");
    printf("  -n pkts num to tx, default = -1,continously\n");
    printf("                default = get pkts from all streams\n");
    printf("  -p print pkts and headers\n");
    printf("  -o print port info, from which pkts come through\n");
    printf("  -f show port filter info \n");
    printf("  -t compare content\n");
    printf("\n");
}

///@brief 打印收包信息的线程，每隔1秒显示一次
///@param ptr 线程建立时的指针，这里没有使用
void print_speed_thread(void *ptr)
{
    ulong old_pkts;
    ulong old_bytes;
    ulong old_tcp_hits;
    ulong old_udp_hits;

    old_pkts = 0;
    old_bytes = 0;
    old_tcp_hits = 0;
    old_udp_hits = 0;

    while (glob_loop_flag_thread)
    {
        sleep(1);

        printf("\nCurrent speed: %8lu pps, %8lu M bps, %8lu tcp hits bps, %8lu udp hits bps\n", glob_pkts - old_pkts,
            (glob_bytes - old_bytes) >> 17,
            glob_tcp_hits - old_tcp_hits,
            glob_udp_hits - old_udp_hits);

            old_pkts = glob_pkts;
            old_bytes = glob_bytes;
            old_tcp_hits = glob_tcp_hits;
            old_udp_hits = glob_udp_hits;
    }
}

///@brief 打印报文（检查收包内容）
///@param pkt 指向报文的指针
///@param len 报文的长度
void compare_payload_content(void *pkt, int len)
{
    int i;
    __u8 *p8;
    static __u64 pkts_num;
    struct ip *pip;
    struct tcphdr *ptcp;
    
    pkts_num++;
    pip = (struct ip *)pkt;
    if(pip->ip_p == IPPROTO_TCP)
    {
        ptcp = (struct tcphdr *)((void *)pip + (pip->ip_hl) * 4);
        p8 =  (__u8 *)((void *)ptcp + ptcp->doff * 4);
        
        len -= (pip->ip_hl * 4 + ptcp->doff * 4);
        ///比较负载内容
        for ( i = 0; i < len; i++)
        {
            if(p8[i] != (__u8)i)
            {
                printf("pkts = %d, len = %d, i = %d:%x, payload = %d\n", pkts_num, len, i, i, p8[i]);
                exit(0);
            }
        }
    }
}

///@brief 打印报文（检查收包内容）
///@param packet_header 指向报文头的指针
///@param packet_content 报文内容指针
void print_pkt(const struct pcap_pkthdr *packet_header, const u_char* packet_content)
{
    int i;

    printf("======= %d pkt === len = %lx:%d ============\n", \
           glob_pkts, packet_header->len, packet_header->len);
    printf("len = %x, caplen = %x\n", packet_header->len, packet_header->caplen);
    printf("sec = %x, usec = %x\n", packet_header->ts.tv_sec, packet_header->ts.tv_usec);
    ///打印报文, 每16字节一行，128字节一段
    for(i = 0; i < packet_header->len; i++)
    {
        if ((i % 128 == 0))
        {
            printf("\n");
        }
        if ((i % 16 == 0))
        {
            printf("\n%03x: ", i / 16 + 1);
        }
        if (i % 8 == 0)
        printf(" ");
        printf("%02x ", packet_content[i]);
    }
    printf("\n");
}
///@brief 处理报文的回调函数,累加收包数，字节数和命中指定端口的包数
void callbak_process_pkt(u_char *argument, const struct pcap_pkthdr *packet_header, const u_char* packet_content)
{
    struct iphdr *iph;
    struct tcphdr *tcph;
    int port_come_through;

    glob_pkts++;

    if (glob_print_pkt_flag)
    {
        print_pkt(packet_header, packet_content);
    }

    if (glob_print_port_flag)
    {
        port_come_through = *(uint8_t *)(packet_content - 2);
        printf("current pkt come from %d port.\n", port_come_through);
    }

    iph = (struct iphdr *)(packet_content + 14);
    glob_bytes += ntohs(iph->tot_len);
    
    
    if (!glob_filter_flg)
    {
        return;
    }

    if(iph->protocol != 0x6 && iph->protocol != 0x11)
    {
        printf("iph->protocol = %d\n", iph->protocol);
        return;
    }

    if (iph->protocol == 0x6)
    {
        //tcph = (struct tcphdr *)(packet_content + 34);
        //if(ntohs(tcph->source) == SPEC_PORT || ntohs(tcph->dest) == SPEC_PORT)
        //{
            glob_tcp_hits++;
        //}
    }
    else
    {
        glob_udp_hits++;
    }
    
    if (glob_compare_content_flag == 1)
    {
        compare_payload_content((void *)iph, ntohs(iph->tot_len));
    }
}

///@brief 信号响应函数，收到信号后把全局循环标志无效掉
void stop_loop(int sig)
{
    glob_loop_flag_thread = 0;
    pcap_breakloop(glob_pcap_handle);
}

///@brief 主函数
///@return 成功返回0，失败返回-1
int main(int argc, char *argv[])
{
    int c, cnt, dev_flg = 0;
    pthread_t print_tid;
    char error_content[PCAP_ERRBUF_SIZE];
    char *dev_name;
    sigset_t set;

    ///读取参数
    cnt = -1;
    while ((c = getopt(argc, argv, "n:i:tpof?")) != -1)
    {
        switch (c)
        {
            case 'n':
                cnt = atoi(optarg);
                break;
            case 'i':
                dev_name = optarg;
                dev_flg = 1;
                break;
            case 't':
                glob_compare_content_flag = 1;
                break;
            case 'p':
                glob_print_pkt_flag = 1;
                break;
            case 'o':
                glob_print_port_flag = 1;
                break;
            case 'f':
                glob_filter_flg = 1;
                break;
            case '?':
            default:
                usage(argv[0]);
                return 0;
        }
    }

    if(!dev_flg)
    {
        usage(argv[0]);
        return 0;
    }
    ///创建打印速度的线程
    glob_loop_flag_thread = 1;
    if (pthread_create(&print_tid, NULL, (void *)print_speed_thread, NULL) != 0)
    {
        printf("### Error: Can not creat threads!\n");
        return -1;
    }

    ///等待结束信号，或着收到期望的包数 
    sigemptyset(&set);
    sigprocmask(SIG_SETMASK, &set, NULL);
    signal(SIGINT, stop_loop);
    
    ///打开设备(BUFSIZ = 8192)
    if((glob_pcap_handle = pcap_open_live(dev_name, BUFSIZ, 1, 0, error_content)) == NULL)
    {
        printf("Erorr: pcap_open_live open failed;dev_name = %s!\n", dev_name);
        return -1;
    }

    ///设置收取带mac头的ethernet包
    pcap_set_datalink(glob_pcap_handle, DLT_EN10MB);

    ///循环收包，用回调函数处理每个包
    pcap_loop(glob_pcap_handle, cnt, callbak_process_pkt, NULL);

    ///停止打印速度的线程
    glob_loop_flag_thread = 0;
    pthread_join(print_tid, NULL);

    ///关闭设备
    pcap_close(glob_pcap_handle);

    ///打印结果
    printf("\n Total : %lu pkts, %lu bytes, %lu tcp hits, %lu udp hits\n", glob_pkts, glob_bytes, glob_tcp_hits, glob_udp_hits);

    return 0;
}
